////////////////////////
// experiment with reference-counted pointers
//
// Beware: this version is presumably not thread-safe.
//   You could make it thread-safe using mutexes.
// Beware: the plain constructor leaves the payload
//   un-initialized, as you can demonstrate in the
//   case where "Payload_t" is "int".

// The "dump" statements are there for pedagogical
// purposes, so that people can learn how reference-
// counted pointers work.

#ifndef PTRCOUNT__H
#define PTRCOUNT__H 1

#include <iostream>

// Forward reference:
template<class Payload_t> class ptrblob;

// A "guts" object contains:
// the uncopyable count, and
// a pointer to the uncopyable payload
template<class Payload_t> class ptrblob_guts {
friend class ptrblob<Payload_t>;
  int refcount;
public:
  Payload_t* loadp;      // pointer to payload

// Constructor:
  ptrblob_guts(int const _refcount, Payload_t * _loadp)
  : refcount(_refcount), loadp(_loadp)
  {}
};

#ifdef DEBUG_PTRCOUNT
#  define dump(x) do_dump(x)
#else
#  define dump(x) /* nothing */
#endif

template<class Payload_t> class ptrblob{
  ptrblob_guts<Payload_t>* guts;
public:

  inline void do_dump(const char* msg){
    unsigned long int addr;
    addr = (unsigned long int) guts;
    std::cout << "\tPtrblob(" << addr << ") "
        << msg << guts->refcount
        << "\tloadp: " << guts->loadp << endl;
  }

// plain constructor
  inline ptrblob() :
    guts(new ptrblob_guts<Payload_t>(0, 0))
  {
    guts->refcount = 1;
    dump("plain constructor:   ");
  };

// copy constructor:
  inline ptrblob(const ptrblob<Payload_t>& old) :
    guts(old.guts)
  {
    guts->refcount++;
    dump("copy_constructor, up to: ");
  }

  inline void decrement() {
    if (guts->refcount <=0) cerr << "counting error" << endl;
    guts->refcount--;
    dump("decrement, down to: ");
    if (guts->refcount <= 0) {
      if (guts->loadp) delete guts->loadp;
      delete guts;
    }
  }

// assignment operator:
  inline ptrblob& operator= (ptrblob const & other) {
    other.guts->refcount++;
    decrement();                // release grep on our previous guts
    guts = other.guts;
    dump("assignment, up to:   ");
    return *this;
  }

// destructor
  inline ~ptrblob(){
    decrement();
  }

//////////// Facilitate access to loadp:
// typical usage:  getsome() = new cmdGetter(_argc, _argv);
  inline Payload_t*& operator() (){
    return guts->loadp;
  }

  inline Payload_t const* operator() () const{
    return (Payload_t const*) guts->loadp;
  }


// Make ptrblob act like a plain old pointer to the payload:
  inline Payload_t& operator* (){
    return *guts->loadp;
  }

  inline Payload_t* operator-> (){
    return guts->loadp;
  }

};

#undef dump
#endif /* PTRCOUNT__H */
